//
// Mikael Klasson
// mklasson@acm.org
// http://mklasson.cjb.net
//
// Finds the function names that a dll exports.
// I've implemented two ways of doing it. 
//   The first is via direct parsing of the exports section of the dll file.
//   The other is via Microsoft's BINDUMP utility.
// Direct dll parsing finds 60 or so functions in kernel32.dll that BINDUMP 
//   doesn't (I guess they weren't in the lib file I used). BINDUMP on the 
//   other hand finds all *Vlm functions, but seeing as how they're not 
//   exported by name in kernel32.dll you'll have trouble intercepting them 
//   anyway. Just go with direct parsing...
//
// 2003-01-12, Fluff, Initial version.
// 2003-10-04, Fluff, Removed the kernel32 hardcoding and made it work with 
//                    any dll.

#define	WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <iostream>
#include <fstream>
#include <string>
#include <vector>
#include <algorithm>
#include <cstdio>
#include "peheader.h"
using namespace std;

int use_dll_directly( string dll_name ) {
	FILE *		fin;
	ofstream	sout( "functions.txt" );
	char		apath[260];
	unsigned int	upeoffset;
	unsigned int	uoffset;
	unsigned int	udata;
	unsigned int	uexprva;
	unsigned int	uexpoffset;
	vector<string>	symbols;
	string		name;

	if( !GetModuleHandle( dll_name.c_str() ) 
		|| !GetModuleFileName( GetModuleHandle( dll_name.c_str() ), apath, sizeof( apath ) ) ) {
		cout << "ERROR: couldn't find " << dll_name << endl;
		return 1;
	}

	cout << "Parsing " << apath << endl;
	if( !( fin = fopen( apath, "rb" ) ) ) {
		cout << "ERROR: opening " << apath << endl;
		return 1;
	}

	fseek( fin, 0x3c, SEEK_SET );
	fread( &upeoffset, 4, 1, fin );
	fseek( fin, upeoffset, SEEK_SET );
	fread( &udata, 4, 1, fin );
	if( udata != 0x004550 ) {
		cout << "ERROR: no PE header" << endl;
		return 1;
	}

	fseek( fin, upeoffset + 15*8, SEEK_SET );
	fread( &uexprva, 4, 1, fin );

	//printf( "Export rva: 0x%X\n", uexprva );
	
	if( !( uexpoffset = rva2phys( fin, uexprva ) ) ) {
		cout << "ERROR: couldn't locate export table" << endl;
		return 1;
	}
	//printf( "Export offset: 0x%X\n", uexpoffset );

	unsigned int	unamecnt,unameptrtabrva,unameptrtaboffset;
	fseek( fin, uexpoffset + 24, SEEK_SET );
	fread( &unamecnt, 4, 1, fin );
	fseek( fin, 4, SEEK_CUR );
	fread( &unameptrtabrva, 4, 1, fin );
	if( !( unameptrtaboffset = rva2phys( fin, unameptrtabrva ) ) ) {
		cout << "ERROR: couldn't locate export name ptr table" << endl;
		return 1;
	}
	for( unsigned int j = 0; j < unamecnt; j++ ) {
		fseek( fin, unameptrtaboffset + j * 4, SEEK_SET );
		fread( &udata, 4, 1, fin );
		uoffset = rva2phys( fin, udata );
		fseek( fin, uoffset, SEEK_SET );
		name = "";
		int	c;
		for( ; ( c = fgetc( fin ) ) != 0; ) {
			name += c;
		}
		symbols.push_back( name );
	}

	sort( symbols.begin(), symbols.end() );
	copy( symbols.begin(), symbols.end(), ostream_iterator<string>( sout, "\n" ) );

	fclose( fin );
	sout.close();

	cout << "Ok. Parsed " << symbols.size() << " symbols and wrote functions.txt." << endl;
	return 0;
}

// This one needs microsofts DUMPBIN utility, and a .lib file
int use_dumpbin( string lib_name ) {
	system( string( "dumpbin /EXPORTS " + lib_name + " > dll_exports" ).c_str() );

	ifstream	sin( "dll_exports" );
	ofstream	sout( "functions.txt" );
	string		name;
	bool		output = false;
	vector<string>	symbols;
	
	for( ; ; ) {
		sin >> name;
		if( name == "Summary" ) {
			break;
		} else if( output ) {
			name = name.substr( 1 );
			name = name.substr( 0, name.find( "@", 0 ) );
			symbols.push_back( name );
		} else if( name == "name" ) {
			output = true;
		}
	}

	sort( symbols.begin(), symbols.end() );
	copy( symbols.begin(), symbols.end(), ostream_iterator<string>( sout, "\n" ) );

	sin.close();
	sout.close();

	cout << "Ok. Parsed " << symbols.size() << " symbols and wrote functions.txt." << endl;
	return 0;
}

int main( int argc, char ** argv ) {
	if( argc < 2 ) {
		cout << "No dll specified. Exiting." << endl;
		return 1;
	}
	//return use_dumpbin( "kernel32.lib" );
	return use_dll_directly( argv[1] );
}
